#include "Eigen/Dense"
#include <iostream>
#include <cmath>

template<typename T>
T f(const T &p, const T &x) { return pow(p*x,2); }

template<typename T>
T dfdp(const T &p, const T &x) { return 2*p*pow(x,2); }

template<typename T>
T ddfdpp(const T&, const T &x) { return 2*pow(x,2); }

template<typename T, int M>
Eigen::Matrix<T,M,1> F(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  auto m=x.size();
  Eigen::Matrix<T,M,1> r(m);
  for (int i=0;i<m;i++) r(i)=f(p,x(i))-y(i);
  return r;
}

template<typename T, int M>
Eigen::Matrix<T,M,1> dFdp(const T &p, const Eigen::Matrix<T,M,1> &x) {
  auto m=x.size();
  Eigen::Matrix<T,M,1> drdp(m);
  for (int i=0;i<m;i++) drdp(i)=dfdp(p,x(i));
  return drdp;
}

template<typename T, int M>
T E(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  return F(p,x,y).squaredNorm();
}

template<typename T, int M>
T dEdp(const T &p, const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y) {
  return 2*F(p,x,y).dot(dFdp(p,x));
}

template<typename T, int M>
T Givens(Eigen::Matrix<T,M,1> a, Eigen::Matrix<T,M,1> y) {
  auto m=a.size();
  for (auto i=m-2;i>=0;i--) {
    T norm_a_tilde=a.block(i,0,2,1).norm();
    y(i)=(a(i)*y(i)+a(i+1)*y(i+1))/norm_a_tilde;
    y(i+1)=(-a(i+1)*y(i)+a(i)*y(i+1))/norm_a_tilde;
    a(i)=norm_a_tilde; a(i+1)=0;
  }
  return y(0)/a(0);
}

template<typename T, int M>
T Regression(const Eigen::Matrix<T,M,1> &x, const Eigen::Matrix<T,M,1> &y, const T& eps) { 
  T p=1;
  while (fabs(dEdp(p,x,y))>eps) {
    Eigen::Matrix<T,M,1> r=F(p,x,y), drdp=dFdp(p,x);
    p-=Givens(drdp,r);
  }
  return p;
}

int main(int argc, char* argv[]) {
  assert(argc==2); auto M=std::stoi(argv[1]);
  using T=double; 
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  VT x=VT::Random(M), y=VT::Random(M); 
  x=x.cwiseProduct(x);
  std::cout << "x^T=" << x.transpose() << std::endl
            << "y^T=" << y.transpose() << std::endl;
  T p=Regression(x,y,1e-5);
  std::cout << "p=" << p << std::endl;
  return 0;
}
